@@ -32,10 +32,11 @@ import inspect
 import traceback
 import os
 import pyatspi
+import re
 import subprocess
 import sys
 
-from . import orca_state
+from datetime import datetime
 
 # Used to turn off all debugging.
 #
@@ -46,7 +47,7 @@ LEVEL_OFF = 10000
 #
 LEVEL_SEVERE = 1000
 
-# Used to decribe events of interest to end users or system managers or which
+# Used to describe events of interest to end users or system managers or which
 # indicate potential problems, but which Orca can deal with without crashing.
 #
 LEVEL_WARNING = 900
@@ -63,7 +64,7 @@ LEVEL_INFO = 800
 # problems that may be associated with a particular configuration.
 #
 # For the purposes of Orca, LEVEL_CONFIGURATION means display the various
-# apsects of whether a particular feature (e.g., speech, braille, etc.)
+# aspects of whether a particular feature (e.g., speech, braille, etc.)
 # is enabled or not as well as details about that feature.
 #
 LEVEL_CONFIGURATION = 700
@@ -118,6 +119,12 @@ debugFile = None
 eventDebugLevel  = LEVEL_FINEST
 eventDebugFilter = None
 
+# If True, we output debug information for the event queue.  We
+# use this in addition to log level to prevent debug logic from
+# bogging down event handling.
+#
+debugEventQueue = False
+
 # What module(s) should be traced if traceit is being used. By default
 # we'll just attend to ourself. (And by default, we will not enable
 # traceit.) Note that enabling this functionality will drag your system
@@ -163,6 +170,8 @@ TRACE_ROLES = []
 #
 TRACE_ONLY_PROCESSING_EVENTS = True
 
+objEvent = None
+
 def printException(level):
     """Prints out information regarding the current exception.
 
@@ -187,7 +196,7 @@ def printStack(level):
         traceback.print_stack(None, 100, debugFile)
         println(level)
 
-def println(level, text = ""):
+def println(level, text="", timestamp=False):
     """Prints the text to stderr unless debug is enabled.
 
     If debug is enabled the text will be redirected to the
@@ -199,6 +208,10 @@ def println(level, text = ""):
     """
 
     if level >= debugLevel:
+        text = text.replace("\ufffc", "[OBJ]")
+        if timestamp:
+            text = text.replace("\n", "\n%s" % (" " * 18))
+            text = "%s - %s" % (datetime.now().strftime("%H:%M:%S.%f"), text)
         if debugFile:
             try:
                 debugFile.writelines([text, "\n"])
@@ -241,7 +254,7 @@ def printResult(level, result=None):
     string = '%s\n%s %s' % (callString, 'RESULT:', result)
     println(level, '%s' % string)
 
-def printObjectEvent(level, event, sourceInfo=None):
+def printObjectEvent(level, event, sourceInfo=None, timestamp=False):
     """Prints out an Python Event object.  The given level may be
     overridden if the eventDebugLevel is greater.  Furthermore, only
     events with event types matching the eventDebugFilter regular
@@ -258,14 +271,18 @@ def printObjectEvent(level, event, sourceInfo=None):
 
     level = max(level, eventDebugLevel)
 
-    text = "OBJECT EVENT: %-40s detail=(%d,%d,%s)" \
-           % (event.type, event.detail1, event.detail2, event.any_data)
-    println(level, text)
+    anydata = event.any_data
+    if isinstance(anydata, str) and len(anydata) > 100:
+        anydata = "%s (...)" % anydata[0:100]
+
+    text = "OBJECT EVENT: %s (%d, %d, %s)" \
+           % (event.type, event.detail1, event.detail2, anydata)
+    println(level, text, timestamp)
 
     if sourceInfo:
-        println(level, "             %s" % sourceInfo)
+        println(level, "%s%s" % (' ' * 18, sourceInfo), timestamp)
 
-def printInputEvent(level, string):
+def printInputEvent(level, string, timestamp=False):
     """Prints out an input event.  The given level may be overridden
     if the eventDebugLevel (see setEventDebugLevel) is greater.
 
@@ -274,9 +291,9 @@ def printInputEvent(level, string):
     - string: the string representing the input event
     """
 
-    println(max(level, eventDebugLevel), string)
+    println(max(level, eventDebugLevel), string, timestamp)
 
-def printDetails(level, indent, accessible, includeApp=True):
+def printDetails(level, indent, accessible, includeApp=True, timestamp=False):
     """Lists the details of the given accessible with the given
     indentation.
 
@@ -289,7 +306,51 @@ def printDetails(level, indent, accessible, includeApp=True):
 
     if level >= debugLevel and accessible:
         println(level,
-                getAccessibleDetails(level, accessible, indent, includeApp))
+                getAccessibleDetails(level, accessible, indent, includeApp),
+                timestamp)
+
+def statesToString(acc, indent=""):
+    try:
+        states = acc.getState().getStates()
+    except:
+        return "%sstates=(exception)" % indent
+
+    return "%sstates='%s'" % (indent, " ".join(map(pyatspi.stateToString, states)))
+
+def relationsToString(acc, indent=""):
+    try:
+        relations = [r.getRelationType() for r in acc.getRelationSet()]
+    except:
+        return "%srelations=(exception)" % indent
+
+    return "%srelations='%s'" % (indent, " ".join(map(pyatspi.relationToString, [r for r in relations])))
+
+def interfacesToString(acc, indent=""):
+    try:
+        interfaces = pyatspi.listInterfaces(acc)
+    except:
+        return "%sinterfaces=(exception)" % indent
+
+    return "%sinterfaces='%s'" % (indent, " ".join(interfaces))
+
+def attributesToString(acc, indent=""):
+    try:
+        attributes = acc.getAttributes()
+    except:
+        return "%sattributes=(exception)" % indent
+
+    return "%sattributes='%s'" % (indent, re.sub("\s+", " ", ", ".join(attributes)))
+
+def actionsToString(acc, indent=""):
+    try:
+        action = acc.queryAction()
+        names = [action.getName(i).lower() for i in range(action.nActions)]
+    except NotImplementedError:
+        return "%sactions=(not implemented)" % indent
+    except:
+        return "%sactions=(exception)" % indent
+
+    return "%sactions='%s'" % (indent, " ".join(names))
 
 def getAccessibleDetails(level, acc, indent="", includeApp=True):
     """Returns a string, suitable for printing, that describes the
@@ -305,57 +366,57 @@ def getAccessibleDetails(level, acc, indent="", includeApp=True):
         return ""
 
     if includeApp:
-        app = acc.getApplication()
-        if app:
-            try:
-                string = indent + "app.name='%s' " % app.name
-            except (LookupError, RuntimeError):
-                string = indent + "app.name='<error getting name>' "
+        try:
+            app = acc.getApplication()
+        except:
+            string = indent + "app=(exception getting app) "
+            app = None
         else:
-            string = indent + "app=None "
+            if app:
+                try:
+                    string = indent + "app.name='%s' " % app.name
+                except (LookupError, RuntimeError):
+                    string = indent + "app.name='(exception getting name)' "
+            else:
+                string = indent + "app=None "
     else:
         string = indent
 
-    # create the States string
     try:
-        stateSet = acc.getState()
+        name_string = "name='%s'".replace("\n", "\\n") % acc.name
     except:
-        string += "(exception getting state set)"
-    try:
-        states = stateSet.getStates()
-    except:
-        string += "(exception getting states)"
-        states = []
-    state_strings = []
-    for state in states:
-        state_strings.append(pyatspi.stateToString(state))
-    state_string = ' '.join(state_strings)
-
-    # create the relations string
-    try:
-        relations = acc.getRelationSet()
-    except:
-        string += "(exception getting relation set)"
-        relations = None
-    if relations:
-        relation_strings = []
-        for relation in relations:
-            relation_strings.append( \
-                          pyatspi.relationToString(relation.getRelationType()))
-        rel_string = ' '.join(relation_strings)
-    else:
-        rel_string = ''
+        name_string = "name=(exception)"
 
     try:
-        string += "name='%s' role='%s' state='%s' relations='%s'" \
-                  % (acc.name or 'None', acc.getRoleName(),
-                     state_string, rel_string)
+        desc_string = "%sdescription='%s'".replace("\n", "\\n") % (indent, acc.description)
+    except:
+        desc_string = "%sdescription=(exception)" % indent
+
+    try:
+        role_string = "role='%s'" % acc.getRoleName()
+    except:
+        role_string = "role=(exception)"
+
+    try:
+        path_string = "%spath=%s" % (indent, pyatspi.getPath(acc))
+    except:
+        path_string = "%spath=(exception)" % indent
+
+    state_string = statesToString(acc, indent)
+    rel_string = relationsToString(acc, indent)
+    actions_string = actionsToString(acc, indent)
+    iface_string = interfacesToString(acc, indent)
+    attr_string = attributesToString(acc, indent)
+
+    try:
+        string += "%s %s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n" \
+                  % (name_string, role_string, desc_string, state_string, rel_string,
+                     actions_string, iface_string, attr_string, path_string)
     except:
         string += "(exception fetching data)"
 
     return string
 
-
 # The following code originated from the following URL:
 # 
 # http://www.dalkescientific.com/writings/diary/archive/ \
@@ -377,7 +438,6 @@ def _getFileAndModule(frame):
     return filename, module
 
 def _shouldTraceIt():
-    objEvent = orca_state.currentObjectEvent
     if not objEvent:
         return not TRACE_ONLY_PROCESSING_EVENTS
 
@@ -462,7 +522,7 @@ def getOpenFDCount(pid):
 
 def getCmdline(pid):
     try:
-        openFile = os.popen('cat /proc/%s/cmdline' % pid)
+        openFile = os.popen('ps -o args -p %s | sed 1d' % pid)
         cmdline = openFile.read()
         openFile.close()
     except:
@@ -479,9 +539,14 @@ def pidOf(procName):
     openFile.close()
     return [int(p) for p in pids.split()]
 
-def examineProcesses():
+def examineProcesses(force=False):
+    if force:
+        level = LEVEL_OFF
+    else:
+        level = LEVEL_ALL
+
     desktop = pyatspi.Registry.getDesktop(0)
-    println(LEVEL_ALL, 'INFO: Desktop has %i apps:' % desktop.childCount)
+    println(level, 'INFO: Desktop has %i apps:' % desktop.childCount, True)
     for i, app in enumerate(desktop):
         pid = app.get_process_id()
         cmd = getCmdline(pid)
@@ -493,19 +558,19 @@ def examineProcesses():
         else:
             if name == '':
                 name = 'WARNING: Possible hang'
-        println(LEVEL_ALL, '%3i. %s (pid: %s) %s file descriptors: %i' \
-                    % (i+1, name, pid, cmd, fds))
+        println(level, '%3i. %s (pid: %s) %s file descriptors: %i' \
+                    % (i+1, name, pid, cmd, fds), True)
 
     # Other 'suspect' processes which might not show up as accessible apps.
     otherApps = ['apport']
     for app in otherApps:
         pids = pidOf(app)
         if not pids:
-            println(LEVEL_ALL, 'INFO: no pid for %s' % app)
+            println(level, 'INFO: no pid for %s' % app, True)
             continue
 
         for pid in pids:
             cmd = getCmdline(pid)
             fds = getOpenFDCount(pid)
-            println(LEVEL_ALL, 'INFO: %s (pid: %s) %s file descriptors: %i' \
-                        % (app, pid, cmd, fds))
+            println(level, 'INFO: %s (pid: %s) %s file descriptors: %i' \
+                        % (app, pid, cmd, fds), True)
