@@ -90,7 +90,6 @@ class EventManager:
                 self.forceLegacyKeyHandling = True
                 self.activateLegacyKeyHandling()
                 return
-            orca_state.device.event_count = 0
             orca_state.device.key_watcher = orca_state.device.add_key_watcher(self._processNewKeyboardEvent)
             self.newKeyHandlingActive = True
 
@@ -138,6 +137,21 @@ class EventManager:
             if eventType in self._ignoredEvents:
                 self._ignoredEvents.remove(eventType)
 
+    def _isDuplicateEvent(self, event):
+        """Returns True if this event is already in the event queue."""
+
+        isSame = lambda x: x.type == event.type \
+            and x.source == event.source \
+            and x.detail1 == event.detail1 \
+            and x.detail2 == event.detail2 \
+            and x.any_data == event.any_data
+
+        for e in self._eventQueue.queue:
+            if isSame(e):
+                return True
+
+        return False
+
     def _ignore(self, event):
         """Returns True if this event should be ignored."""
 
@@ -175,13 +189,19 @@ class EventManager:
             debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
+        if self._isDuplicateEvent(event):
+            msg = 'EVENT MANAGER: Ignoring duplicate event'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         if self._inDeluge() and self._ignoreDuringDeluge(event):
             msg = 'EVENT MANAGER: Ignoring event type due to deluge'
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
         script = orca_state.activeScript
-        if event.type.startswith('object:children-changed'):
+        if event.type.startswith('object:children-changed') \
+           or event.type.startswith('object:state-changed:sensitive'):
             if not script:
                 msg = 'EVENT MANAGER: Ignoring because there is no active script'
                 debug.println(debug.LEVEL_INFO, msg, True)
@@ -258,6 +278,11 @@ class EventManager:
                 msg = 'EVENT MANAGER: Ignoring event type due to role'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
+        elif event.type.startswith('object:state-changed:selected'):
+            if not event.detail1 and role in [pyatspi.ROLE_PUSH_BUTTON]:
+                msg = 'EVENT MANAGER: Ignoring event type due to role and detail1'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
         elif event.type.startswith('object:state-changed:showing'):
             if role not in [pyatspi.ROLE_ALERT,
                             pyatspi.ROLE_ANIMATION,
@@ -271,6 +296,21 @@ class EventManager:
                 msg = 'EVENT MANAGER: Ignoring event type due to role'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
+            if role == pyatspi.ROLE_PANEL:
+                if not event.detail1:
+                    msg = 'EVENT MANAGER: Ignoring event type due to role and detail1'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return True
+                try:
+                    if not event.source.name:
+                        msg = 'EVENT MANAGER: Ignoring event type due to role and lack of name'
+                        debug.println(debug.LEVEL_INFO, msg, True)
+                        return True
+                except:
+                    msg = 'EVENT MANAGER: Ignoring event from dead source'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return True
+
         elif event.type.startswith('object:selection-changed'):
             if event.source in self._parentsOfDefunctDescendants:
                 msg = 'EVENT MANAGER: Ignoring event from parent of defunct descendants'
@@ -346,6 +386,14 @@ class EventManager:
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
 
+            # In normal apps we would have caught this from the parent role.
+            # But gnome-shell has panel parents adding/removing menu items.
+            if event.type.startswith('object:children-changed'):
+                if childRole == pyatspi.ROLE_MENU_ITEM:
+                    msg = 'EVENT MANAGER: Ignoring event type due to child role'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return True
+
         msg = 'EVENT MANAGER: Not ignoring due to lack of cause'
         debug.println(debug.LEVEL_INFO, msg, True)
         return False
@@ -446,10 +494,15 @@ class EventManager:
                 toolkitName = app.toolkitName
             except:
                 toolkitName = None
-            if toolkitName in self._synchronousToolkits \
-               or isinstance(e, input_event.MouseButtonEvent) \
-               or e.type.startswith("object:children-changed"):
+            if isinstance(e, input_event.MouseButtonEvent):
+                asyncMode = True
+            elif toolkitName in self._synchronousToolkits:
                 asyncMode = False
+            elif e.type.startswith("object:children-changed"):
+                try:
+                    asyncMode = e.source.getRole() == pyatspi.ROLE_TABLE
+                except:
+                    asyncMode = True
             script = _scriptManager.getScript(app, e.source)
             script.eventCache[e.type] = (e, time.time())
 
@@ -674,7 +727,7 @@ class EventManager:
         """Returns the script associated with event."""
 
         if event.type.startswith("mouse:"):
-            return orca_state.activeScript
+            return _scriptManager.getScriptForMouseButtonEvent(event)
 
         script = None
         app = None
@@ -736,9 +789,8 @@ class EventManager:
         
         if not script:
             script = self._getScriptForEvent(event)
-
-        if not script:
-            return False, "There is no script for this event."
+            if not script:
+                return False, "There is no script for this event."
 
         if script == orca_state.activeScript:
             return False, "The script for this event is already active."
@@ -781,9 +833,22 @@ class EventManager:
 
         return False, "No reason found to activate a different script."
 
+    def _eventSourceIsDead(self, event):
+        try:
+            name = event.source.name
+        except:
+            msg = "EVENT MANAGER: source of %s is dead" % event.type
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        return False
+
     def _ignoreDuringDeluge(self, event):
         """Returns true if this event should be ignored during a deluge."""
 
+        if self._eventSourceIsDead(event):
+            return True
+
         ignore = ["object:text-changed:delete",
                   "object:text-changed:insert",
                   "object:text-changed:delete:system",
@@ -813,6 +878,9 @@ class EventManager:
     def _processDuringFlood(self, event):
         """Returns true if this event should be processed during a flood."""
 
+        if self._eventSourceIsDead(event):
+            return False
+
         ignore = ["object:text-changed:delete",
                   "object:text-changed:insert",
                   "object:text-changed:delete:system",
@@ -1026,10 +1094,10 @@ class EventManager:
         event.event_string = text
         if event.event_string is None:
             event.event_string = ""
-        event.timestamp = device.event_count
-        device.event_count = device.event_count + 1
+        event.timestamp = time.time()
 
-        if not pressed and text == "Num_Lock" and "KP_Insert" in settings.orcaModifierKeys and orca_state.activeSWcript is not None:
+        if not pressed and text == "Num_Lock" and "KP_Insert" in settings.orcaModifierKeys \
+            and orca_state.activeScript is not None:
             orca_state.activeScript.refreshKeyGrabs()
 
         if pressed:
