@@ -55,30 +55,30 @@ class LabelInference:
         Returns the text which we think is the label, or None.
         """
 
-        debug.println(debug.LEVEL_FINE, "INFER label for: %s" % obj)
+        debug.println(debug.LEVEL_INFO, "INFER label for: %s" % obj, True)
         if not obj:
             return None, []
 
         if focusedOnly and not obj.getState().contains(pyatspi.STATE_FOCUSED):
-            debug.println(debug.LEVEL_FINE, "INFER - object not focused")
+            debug.println(debug.LEVEL_INFO, "INFER - object not focused", True)
             return None, []
 
         result, objects = None, []
         if not result:
             result, objects = self.inferFromTextLeft(obj)
-            debug.println(debug.LEVEL_FINE, "INFER - Text Left: %s" % result)
+            debug.println(debug.LEVEL_INFO, "INFER - Text Left: %s" % result, True)
         if not result or self._preferRight(obj):
             result, objects = self.inferFromTextRight(obj) or result
-            debug.println(debug.LEVEL_FINE, "INFER - Text Right: %s" % result)
+            debug.println(debug.LEVEL_INFO, "INFER - Text Right: %s" % result, True)
         if not result:
             result, objects = self.inferFromTable(obj)
-            debug.println(debug.LEVEL_FINE, "INFER - Table: %s" % result)
+            debug.println(debug.LEVEL_INFO, "INFER - Table: %s" % result, True)
         if not result:
             result, objects = self.inferFromTextAbove(obj)
-            debug.println(debug.LEVEL_FINE, "INFER - Text Above: %s" % result)
+            debug.println(debug.LEVEL_INFO, "INFER - Text Above: %s" % result, True)
         if not result:
             result, objects = self.inferFromTextBelow(obj)
-            debug.println(debug.LEVEL_FINE, "INFER - Text Below: %s" % result)
+            debug.println(debug.LEVEL_INFO, "INFER - Text Below: %s" % result, True)
 
         # TODO - We probably do not wish to "infer" from these. Instead, we
         # should ensure that this content gets presented as part of the widget.
@@ -86,14 +86,16 @@ class LabelInference:
         # are each something other than a label.)
         if not result:
             result, objects = obj.name, []
-            debug.println(debug.LEVEL_FINE, "INFER - Name: %s" % result)
-        if not result:
-            result, objects = obj.description, []
-            debug.println(debug.LEVEL_FINE, "INFER - Description: %s" % result)
+            debug.println(debug.LEVEL_INFO, "INFER - Name: %s" % result, True)
         if result:
             result = result.strip()
             result = result.replace("\n", " ")
 
+        # Desperate times call for desperate measures....
+        if not result:
+            result, objects = self.inferFromTextLeft(obj, proximity=200)
+            debug.println(debug.LEVEL_INFO, "INFER - Text Left with proximity of 200: %s" % result, True)
+
         self.clearCache()
         return result, objects
 
@@ -147,10 +149,12 @@ class LabelInference:
         if not obj:
             return False
 
+        isMatch = lambda x: x and not self._script.utilities.isStaticTextLeaf(x)
+
         try:
-            children = [child for child in obj]
+            children = [child for child in obj if isMatch(child)]
         except (LookupError, RuntimeError):
-            debug.println(debug.LEVEL_FINE, 'Dead Accessible in %s' % obj)
+            debug.println(debug.LEVEL_INFO, 'Dead Accessible in %s' % obj, True)
             return False
 
         children = [x for x in children if x.getRole() != pyatspi.ROLE_LINK]
@@ -187,7 +191,7 @@ class LabelInference:
             return False
 
         rv = self._isWidgetCache.get(hash(obj))
-        if rv != None:
+        if rv is not None:
             return rv
 
         widgetRoles = [pyatspi.ROLE_CHECK_BOX,
@@ -203,6 +207,9 @@ class LabelInference:
                        pyatspi.ROLE_PUSH_BUTTON]
 
         isWidget = obj.getRole() in widgetRoles
+        if not isWidget and obj.getState().contains(pyatspi.STATE_EDITABLE):
+            isWidget = True
+
         self._isWidgetCache[hash(obj)] = isWidget
         return isWidget
 
@@ -219,20 +226,31 @@ class LabelInference:
             return rv
 
         extents = 0, 0, 0, 0
-        try:
-            text = obj.queryText()
-        except NotImplementedError:
-            pass
-        else:
+        text = self._script.utilities.queryNonEmptyText(obj)
+        if text:
             skipTextExtents = [pyatspi.ROLE_ENTRY, pyatspi.ROLE_PASSWORD_TEXT]
             if not obj.getRole() in skipTextExtents:
                 if endOffset == -1:
-                    endOffset = text.characterCount
+                    try:
+                        endOffset = text.characterCount
+                    except:
+                        msg = "ERROR: Exception getting character count for %s" % obj
+                        debug.println(debug.LEVEL_INFO, msg, True)
+                        return extents
+
                 extents = text.getRangeExtents(startOffset, endOffset, 0)
 
         if not (extents[2] and extents[3]):
-            ext = obj.queryComponent().getExtents(0)
-            extents = ext.x, ext.y, ext.width, ext.height
+            try:
+                ext = obj.queryComponent().getExtents(0)
+            except NotImplementedError:
+                msg = "INFO: %s does not implement the component interface" % obj
+                debug.println(debug.LEVEL_INFO, msg, True)
+            except:
+                msg = "ERROR: Exception getting extents for %s" % obj
+                debug.println(debug.LEVEL_INFO, msg, True)
+            else:
+                extents = ext.x, ext.y, ext.width, ext.height
 
         self._extentsCache[(hash(obj), startOffset, endOffset)] = extents
         return extents
@@ -241,20 +259,20 @@ class LabelInference:
         """Gets the functional label text associated with the object obj."""
 
         if not self._isSimpleObject(obj):
-            return ''
+            return None, []
 
         if self._cannotLabel(obj):
-            return ''
+            return None, []
 
-        contents = self._script.utilities.getObjectsFromEOCs(obj)
+        contents = self._script.utilities.getObjectContentsAtOffset(obj, useCache=False)
         objects = [content[0] for content in contents]
         if list(filter(self._isWidget, objects)):
-            return ''
+            return None, []
 
-        strings = [content[3] or content[0].name for content in contents]
-        return ''.join(strings)
+        strings = [content[3] for content in contents]
+        return ''.join(strings), objects
 
-    def _getLineContents(self, obj):
+    def _getLineContents(self, obj, start=0):
         """Get the (obj, startOffset, endOffset, string) tuples for the line
         containing the object, obj."""
 
@@ -263,37 +281,15 @@ class LabelInference:
             return rv
 
         key = hash(obj)
-        start = None
         if self._isWidget(obj):
             start, end = self._script.utilities.getHyperlinkRange(obj)
             obj = obj.parent
 
-        try:
-            text = obj.queryText()
-        except:
-            start = 0
-        else:
-            if start == None:
-                start = max(0, text.caretOffset)
-
-        rv = self._script.utilities.getLineContentsAtOffset(obj, start)
+        rv = self._script.utilities.getLineContentsAtOffset(obj, start, True, False)
         self._lineCache[key] = rv
 
         return rv
 
-    def _getPreviousObject(self, obj):
-        """Gets the object prior to obj."""
-
-        index = obj.getIndexInParent()
-        if not index > 0:
-            return obj.parent
-
-        prevObj = obj.parent[index-1]
-        if prevObj and prevObj.childCount:
-            prevObj = prevObj[prevObj.childCount - 1]
-
-        return prevObj
-
     def inferFromTextLeft(self, obj, proximity=75):
         """Attempt to infer the functional/displayed label of obj by
         looking at the contents of the current line, which are to the
@@ -331,9 +327,9 @@ class LabelInference:
         lExtents = self._getExtents(lObj, start, end)
         distance = extents[0] - (lExtents[0] + lExtents[2])
         if 0 <= distance <= proximity:
-            strings = [content[3] or content[0].name for content in onLeft]
-            result = ''.join(strings)
-            if result.strip():
+            strings = [content[3] for content in onLeft]
+            result = ''.join(strings).strip()
+            if result:
                 return result, [content[0] for content in onLeft]
 
         return None, []
@@ -378,9 +374,9 @@ class LabelInference:
         rExtents = self._getExtents(rObj, start, end)
         distance = rExtents[0] - (extents[0] + extents[2])
         if distance <= proximity or self._preferRight(obj):
-            strings = [content[3] or content[0].name for content in onRight]
-            result = ''.join(strings)
-            if result.strip():
+            strings = [content[3] for content in onRight]
+            result = ''.join(strings).strip()
+            if result:
                 return result, [content[0] for content in onRight]
 
         return None, []
@@ -398,51 +394,30 @@ class LabelInference:
         """
 
         thisLine = self._getLineContents(obj)
-        if not (thisLine and thisLine[0]):
-            return None, []
-
-        prevObj, start, end, string = thisLine[0]
-        if obj == prevObj:
-            start, end = self._script.utilities.getHyperlinkRange(prevObj)
-            prevObj = prevObj.parent
-
+        content = [o for o in thisLine if o[0] == obj]
         try:
-            text = prevObj.queryText()
-        except (AttributeError, NotImplementedError):
+            index = thisLine.index(content[0])
+        except IndexError:
+            return None, []
+        if index > 0:
             return None, []
 
-        objX, objY, objWidth, objHeight = self._getExtents(obj)
-        if not (objWidth and objHeight):
-            return None, []
-
-        start = max(start - 1, 0)
-        prevLine = self._script.utilities.getLineContentsAtOffset(prevObj, start)
-        if not (prevLine and prevLine[0]):
+        prevObj, prevOffset = self._script.utilities.previousContext(
+            thisLine[0][0], thisLine[0][1], True)
+        prevLine = self._getLineContents(prevObj, prevOffset)
+        if len(prevLine) != 1:
             return None, []
 
         prevObj, start, end, string = prevLine[0]
-        if string.strip() and not self._cannotLabel(prevObj):
-            x, y, width, height = self._getExtents(prevObj, start, end)
-            distance = objY - (y + height)
-            if 0 <= distance <= proximity:
-                return string, [prevObj]
+        if self._cannotLabel(prevObj):
+            return None, []
 
-        while prevObj:
-            prevObj = self._getPreviousObject(prevObj)
-            x, y, width, height = self._getExtents(prevObj)
+        if string.strip():
+            x, y, width, height = self._getExtents(prevObj, start, end)
+            objX, objY, objWidth, objHeight = self._getExtents(obj)
             distance = objY - (y + height)
-            if distance > proximity:
-                return None, []
-            if prevObj.getRole() == pyatspi.ROLE_TABLE_CELL \
-               and not prevObj in [obj.parent, obj.parent.parent]:
-                return None, []
-            if distance < 0:
-                continue
-            if x + 150 < objX:
-                continue
-            string = self._createLabelFromContents(prevObj)
-            if string:
-                return string, [prevObj]
+            if 0 <= distance <= proximity and x <= objX:
+                return string.strip(), [prevObj]
 
         return None, []
 
@@ -462,31 +437,83 @@ class LabelInference:
             return None, []
 
         thisLine = self._getLineContents(obj)
-        if not (thisLine and thisLine[0]):
+        content = [o for o in thisLine if o[0] == obj]
+        try:
+            index = thisLine.index(content[0])
+        except IndexError:
+            return None, []
+        if index > 0:
             return None, []
 
-        lastObj, start, end, string = thisLine[-1]
-        if obj == lastObj:
-            start, end = self._script.utilities.getHyperlinkRange(obj)
-            lastObj = lastObj.parent
-
-        objX, objY, objWidth, objHeight = self._getExtents(obj)
-        if not (objWidth and objHeight):
-            return None, []
-
-        nextLine = self._script.utilities.getLineContentsAtOffset(lastObj, end)
-        if not (nextLine and nextLine[0]):
+        nextObj, nextOffset = self._script.utilities.nextContext(
+            thisLine[-1][0], thisLine[-1][2] - 1, True)
+        nextLine = self._getLineContents(nextObj, nextOffset)
+        if len(nextLine) != 1:
             return None, []
 
         nextObj, start, end, string = nextLine[0]
-        if string.strip() and not self._cannotLabel(nextObj):
+        if self._cannotLabel(nextObj):
+            return None, []
+
+        if string.strip():
             x, y, width, height = self._getExtents(nextObj, start, end)
+            objX, objY, objWidth, objHeight = self._getExtents(obj)
             distance = y - (objY + objHeight)
             if 0 <= distance <= proximity:
-                return string, [nextObj]
+                return string.strip(), [nextObj]
 
         return None, []
 
+    def _isTable(self, obj):
+        if not obj:
+            return False
+
+        if obj.getRole() == pyatspi.ROLE_TABLE:
+            return True
+
+        return self._getTag(obj) == 'table'
+
+    def _isRow(self, obj):
+        if not obj:
+            return False
+
+        if obj.getRole() == pyatspi.ROLE_TABLE_ROW:
+            return True
+
+        return self._getTag(obj) == 'tr'
+
+    def _isCell(self, obj):
+        if not obj:
+            return False
+
+        if obj.getRole() == pyatspi.ROLE_TABLE_CELL:
+            return True
+
+        return self._getTag(obj) in ['td', 'th']
+
+    def _getCellFromTable(self, table, rowindex, colindex):
+        if "Table" not in pyatspi.listInterfaces(table):
+            return None
+
+        if rowindex < 0 or colindex < 0:
+            return None
+
+        iface = table.queryTable()
+        if rowindex >= iface.nRows or colindex >= iface.nColumns:
+            return None
+
+        return table.queryTable().getAccessibleAt(rowindex, colindex)
+
+    def _getCellFromRow(self, row, colindex):
+        if 0 <= colindex < row.childCount:
+            return row[colindex]
+
+        return None
+
+    def _getTag(self, obj):
+        attrs = self._script.utilities.objectAttributes(obj)
+        return attrs.get('tag')
+
     def inferFromTable(self, obj, proximityForRight=50):
         """Attempt to infer the functional/displayed label of obj by looking
         at the contents of the surrounding table cells. Note that this approach
@@ -499,54 +526,58 @@ class LabelInference:
         Returns the text which we think is the label, or None.
         """
 
-        pred = lambda x: x.getRole() == pyatspi.ROLE_TABLE_CELL
-        cell = pyatspi.utils.findAncestor(obj, pred)
+        cell = pyatspi.findAncestor(obj, self._isCell)
         if not self._isSimpleObject(cell):
             return None, []
 
         if not cell in [obj.parent, obj.parent.parent]:
             return None, []
 
-        pred = lambda x: x.getRole() == pyatspi.ROLE_TABLE
-        grid = pyatspi.utils.findAncestor(cell, pred)
+        grid = pyatspi.findAncestor(cell, self._isTable)
         if not grid:
             return None, []
 
-        try:
-            table = grid.queryTable()
-        except NotImplementedError:
-            return None, []
+        cellLeft = cellRight = cellAbove = cellBelow = None
+        gridrow = pyatspi.findAncestor(cell, self._isRow)
+        rowindex, colindex = self._script.utilities.coordinatesForCell(cell)
+        if colindex > -1:
+            cellLeft = self._getCellFromTable(grid, rowindex, colindex - 1)
+            cellRight = self._getCellFromTable(grid, rowindex, colindex + 1)
+            cellAbove = self._getCellFromTable(grid, rowindex - 1, colindex)
+            cellBelow = self._getCellFromTable(grid, rowindex + 1, colindex)
+        elif gridrow and cell.parent == gridrow:
+            cellindex = cell.getIndexInParent()
+            cellLeft = self._getCellFromRow(gridrow, cellindex - 1)
+            cellRight = self._getCellFromRow(gridrow, cellindex + 1)
+            rowindex = gridrow.getIndexInParent()
+            if rowindex > 0:
+                cellAbove = self._getCellFromRow(gridrow.parent[rowindex - 1], cellindex)
+            if rowindex + 1 < grid.childCount:
+                cellBelow = self._getCellFromRow(gridrow.parent[rowindex + 1], cellindex)
+
+        if cellLeft and not self._preferRight(obj):
+            label, sources = self._createLabelFromContents(cellLeft)
+            if label:
+                return label.strip(), sources
 
-        index = self._script.utilities.cellIndex(cell)
-        row = table.getRowAtIndex(index)
-        col = table.getColumnAtIndex(index)
         objX, objY, objWidth, objHeight = self._getExtents(obj)
 
-        if col > 0 and not self._preferRight(obj):
-            candidate = table.getAccessibleAt(row, col - 1)
-            label = self._createLabelFromContents(candidate)
-            if label.strip():
-                return label, [candidate]
-
-        if col < table.nColumns and not self._preventRight(obj):
-            candidate = table.getAccessibleAt(row, col + 1)
-            x, y, width, height = self._getExtents(candidate)
+        if cellRight and not self._preventRight(obj):
+            x, y, width, height = self._getExtents(cellRight)
             distance = x - (objX + objWidth)
             if distance <= proximityForRight or self._preferRight(obj):
-                label = self._createLabelFromContents(candidate)
-                if label.strip():
-                    return label, [candidate]
+                label, sources = self._createLabelFromContents(cellRight)
+                if label:
+                    return label.strip(), sources
 
-        cellAbove = cellBelow = labelAbove = labelBelow = None
-        if row > 0:
-            cellAbove = table.getAccessibleAt(row - 1, col)
-            labelAbove = self._createLabelFromContents(cellAbove)
+        labelAbove = labelBelow = None
+        if cellAbove:
+            labelAbove, sourcesAbove = self._createLabelFromContents(cellAbove)
             if labelAbove and self._preferTop(obj):
-                return labelAbove, [cellAbove]
+                return labelAbove.strip(), sourcesAbove
 
-        if row < table.nRows and not self._preventBelow(obj):
-            cellBelow = table.getAccessibleAt(row + 1, col)
-            labelBelow = self._createLabelFromContents(cellBelow)
+        if cellBelow and not self._preventBelow(obj):
+            labelBelow, sourcesBelow = self._createLabelFromContents(cellBelow)
 
         if labelAbove and labelBelow:
             aboveX, aboveY, aboveWidth, aboveHeight = self._getExtents(cellAbove)
@@ -554,28 +585,37 @@ class LabelInference:
             dAbove = objY - (aboveY + aboveHeight)
             dBelow = belowY - (objY + objHeight)
             if dAbove <= dBelow:
-                return labelAbove, [cellAbove]
-            return labelBelow, [cellBelow]
+                return labelAbove.strip(), sourcesAbove
+            return labelBelow.strip(), sourcesBelow
 
         if labelAbove:
-            return labelAbove, [cellAbove]
+            return labelAbove.strip(), sourcesAbove
         if labelBelow:
-            return labelBelow, [cellBelow]
+            return labelBelow.strip(), sourcesBelow
 
         # None of the cells immediately surrounding this cell seem to be serving
         # as a functional label. Therefore, see if this table looks like a grid
         # of widgets with the functional labels in the first row.
+
+        try:
+            table = grid.queryTable()
+        except NotImplementedError:
+            return None, []
+
         firstRow = [table.getAccessibleAt(0, i) for i in range(table.nColumns)]
         if not firstRow or list(filter(self._isWidget, firstRow)):
             return None, []
 
-        cells = [table.getAccessibleAt(i, col) for i in range(1, table.nRows)]
-        cells = [x for x in cells if x != None]
+        if colindex < 0:
+            return None, []
+
+        cells = [table.getAccessibleAt(i, colindex) for i in range(1, table.nRows)]
+        cells = [x for x in cells if x is not None]
         if [x for x in cells if x.childCount and x[0].getRole() != obj.getRole()]:
             return None, []
 
-        label = self._createLabelFromContents(firstRow[col])
+        label, sources = self._createLabelFromContents(firstRow[colindex])
         if label:
-            return label, [firstRow[col]]
+            return label.strip(), sources
 
         return None, []
