@@ -28,7 +28,7 @@ __license__   = "LGPL"
 
 # $ORCA_VERSION
 #
-version     = "43.1"
+version     = "44.1"
 
 # The revision if built from git; otherwise an empty string
 #
@@ -48,4 +48,4 @@ datadir = "${prefix}/share".replace('${prefix}', '/usr/local')
 
 # The directory where we could find liblouis translation tables.
 #
-tablesdir = "/usr/local/share/liblouis/tables"
+tablesdir = ""
