@@ -82,8 +82,20 @@ class SayAllContext:
         self.obj           = obj
         self.utterance     = utterance
         self.startOffset   = startOffset
-        self.currentOffset = startOffset
         self.endOffset     = endOffset
+        self.currentOffset = startOffset
+        self.currentEndOffset = None
+
+    def __str__(self):
+        return "SAY ALL: %s '%s' (%i-%i, current: %i)" % \
+            (self.obj, self.utterance, self.startOffset, self.endOffset, self.currentOffset)
+
+    def copy(self):
+        new = SayAllContext(self.obj, self.utterance,
+                            self.startOffset, self.endOffset)
+        new.currentOffset = self.currentOffset
+        new.currentEndOffset = self.currentEndOffset
+        return new
 
 
 class SpeechServer(object):
@@ -149,20 +161,6 @@ class SpeechServer(object):
         """
         pass
 
-    def speakUtterances(self, utteranceList, acss=None, interrupt=True):
-        """Speaks the given list of utterances immediately.
-
-        Arguments:
-        - utteranceList: list of strings to be spoken
-        - acss:      acss.ACSS instance; if None,
-                     the default voice settings will be used.
-                     Otherwise, the acss settings will be
-                     used to augment/override the default
-                     voice settings.
-        - interrupt: if True, stop any speech currently in progress.
-        """
-        pass
-
     def speak(self, text=None, acss=None, interrupt=True):
         """Speaks all queued text immediately.  If text is not None,
         it is added to the queue before speaking.
@@ -179,10 +177,6 @@ class SpeechServer(object):
         """
         pass
 
-    def isSpeaking(self):
-        """"Returns True if the system is currently speaking."""
-        return False
-
     def sayAll(self, utteranceIterator, progressCallback):
         """Iterates through the given utteranceIterator, speaking
         each utterance one at a time.  Subclasses may postpone
